/*
 * Copyright 2008 Evan Cowden
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License. 
 * You may obtain a copy of the License at 
 * 
 *      http://www.apache.org/licenses/LICENSE-2.0 
 * 
 * Unless required by applicable law or agreed to in writing, software 
 * distributed under the License is distributed on an "AS IS" BASIS, 
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
 * See the License for the specific language governing permissions and 
 * limitations under the License.
 */
package org.nakedosgi.spi.processors;

import java.util.Set;
import javax.annotation.processing.AbstractProcessor;
import javax.annotation.processing.RoundEnvironment;
import javax.annotation.processing.SupportedAnnotationTypes;
import javax.annotation.processing.SupportedSourceVersion;
import javax.lang.model.SourceVersion;
import javax.lang.model.element.Element;
import javax.lang.model.element.TypeElement;
import javax.lang.model.type.MirroredTypeException;
import javax.lang.model.type.TypeMirror;
import org.nakedosgi.spi.ServiceProviderAnnotation;

/**
 * The annotation processor for the 
 * <code>&#064;</code>{@link ServiceProviderAnnotation} 
 * annotation
 * 
 * @author Evan Cowden
 */
@SupportedAnnotationTypes("org.nakedosgi.spi.ServiceProviderAnnotation")
@SupportedSourceVersion(SourceVersion.RELEASE_6)
public class ServiceProviderAnnotationProcessor extends AbstractProcessor {

    //Map between services and service providers
    private SimpleMultiMap<String, String> configurations = new SimpleMultiMap();

    @Override
    public boolean process(Set<? extends TypeElement> annotations, RoundEnvironment roundEnv) {
        //Find all annotations that have the ServiceProviderAnnotation annotation
        Set<? extends Element> annotatedAnnotations = roundEnv.getElementsAnnotatedWith(ServiceProviderAnnotation.class);
        for (Element annotatedAnnotation : annotatedAnnotations) {
            //Get the servcie name from the meta-annotation
            ServiceProviderAnnotation providerAnnotation = annotatedAnnotation.getAnnotation(ServiceProviderAnnotation.class);
            TypeMirror serviceType = getValue(providerAnnotation);
            String service = Util.getQualifiedName(serviceType);

            //Get all elements annotated with the provider annotation
            Set<? extends Element> elements = roundEnv.getElementsAnnotatedWith((TypeElement) annotatedAnnotation);
            for (Element element : elements) {
                boolean isValid = Util.validateProvider(processingEnv, element, serviceType);
                if (isValid) {
                    String provider = Util.getQualifiedName(element.asType());
                    configurations.put(service, provider);
                }
            }
        }

        //Write if processing is complete
        if (roundEnv.processingOver()) {
            Util.writeConfigurations(processingEnv.getFiler(), configurations);
        }

        return true;
    }

    /**
     * Get the value field out of the {@code ServiceProviderAnnotation} 
     * annotation.
     * @param annotation 
     */
    private static TypeMirror getValue(ServiceProviderAnnotation annotation) {
        //Get the value from the annotation using the fun MirroredTypeException workaround
        TypeMirror result = null;
        try {
            Class thisLineWillThrowAnException = annotation.value();
        } catch (MirroredTypeException ex) {
            result = ex.getTypeMirror();
        }

        return result;
    }
}
