/*
 * Copyright 2008 Evan Cowden
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License. 
 * You may obtain a copy of the License at 
 * 
 *      http://www.apache.org/licenses/LICENSE-2.0 
 * 
 * Unless required by applicable law or agreed to in writing, software 
 * distributed under the License is distributed on an "AS IS" BASIS, 
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
 * See the License for the specific language governing permissions and 
 * limitations under the License.
 */
package org.nakedosgi.spi.processors;

import java.util.List;
import java.util.Set;
import javax.annotation.processing.AbstractProcessor;
import javax.annotation.processing.Messager;
import javax.annotation.processing.RoundEnvironment;
import javax.annotation.processing.SupportedAnnotationTypes;
import javax.annotation.processing.SupportedSourceVersion;
import javax.lang.model.SourceVersion;
import javax.lang.model.element.Element;
import javax.lang.model.element.TypeElement;
import javax.lang.model.type.MirroredTypeException;
import javax.lang.model.type.TypeMirror;
import javax.tools.Diagnostic.Kind;
import org.nakedosgi.spi.ServiceProvider;

/**
 * The annotation processor for the <code>&#064;</code>{@link ServiceProvider} annotation
 * 
 * @author Evan Cowden
 */
@SupportedAnnotationTypes("org.nakedosgi.spi.ServiceProvider")
@SupportedSourceVersion(SourceVersion.RELEASE_6)
public class ServiceProviderProcessor extends AbstractProcessor {

    //Map between services and service providers
    private SimpleMultiMap<String, String> configurations = new SimpleMultiMap();

    @Override
    public boolean process(Set<? extends TypeElement> annotations, RoundEnvironment roundEnv) {
        Set<? extends Element> elements = roundEnv.getElementsAnnotatedWith(ServiceProvider.class);
        for (Element rawElement : elements) {
            TypeElement element = (TypeElement) rawElement;
            String providerName = element.getQualifiedName().toString();

            //Examine the annotation to determine if an interface is set explicitely
            ServiceProvider annotation = element.getAnnotation(ServiceProvider.class);
            TypeMirror value = getValue(annotation);

            String valueText = Util.getQualifiedName(value);
            if (valueText.equals(Void.class.getName())) { //Empty annotation value
                //...auto-detect services
                List<? extends TypeMirror> interfaces = element.getInterfaces();

                if (interfaces.size() == 0) {
                    Messager messager = processingEnv.getMessager();
                    messager.printMessage(Kind.ERROR, String.format("Annotated service provider '%s' specifies zero interfaces.\n  Either explicitely define an interface or set @ServiceProvider annotation value.", providerName), element);
                }

                for (TypeMirror iface : interfaces) {
                    boolean isValid = Util.validateProvider(processingEnv, rawElement, iface);
                    if (isValid) {
                        String service = Util.getQualifiedName(iface);
                        configurations.put(service, providerName);
                    }
                }

            } else { //Annotation value is specified

                boolean isValid = Util.validateProvider(processingEnv, rawElement, value);
                if (isValid) {
                    //...use specified service
                    configurations.put(valueText, providerName);
                }
            }
        }

        //Write configurations when processing is complete
        if (roundEnv.processingOver()) {
            Util.writeConfigurations(processingEnv.getFiler(), configurations);
        }

        return true;
    }

    /**
     * Get the value field out of the ServiceProvider annotation.  Use when the
     * type of the value is a {@code Class}.
     * @param annotation with a <i>value</i> field of type {@code Class}
     */
    private static TypeMirror getValue(ServiceProvider annotation) {
        //Get the value from the annotation using the fun MirroredTypeException workaround
        TypeMirror result = null;
        try {
            Class thisLineWillThrowAnException = annotation.value();
        } catch (MirroredTypeException ex) {
            result = ex.getTypeMirror();
        }

        return result;
    }
}
